:- use_module(normalform).
:- use_module(struct).
:- use_module(ids).
:- use_module(session).
:- use_module(logging).

:-declare_struct(ausw,[posMap,varMap,musterL]).

%ineffizient
selectNth(Element,List,Pos,Rest):-
	append(Rest1,[Element|Rest2],List),
	length(Rest1,Pos),
	append(Rest1,Rest2,Rest).

auswertung(Loesung,AufgabenId,Fehler):-
	logging(auswertung,'<auswertung>Input: Loesung=~w, AufgabenId=~w',[Loesung,AufgabenId]),
	ids_search(auswertung0(AufgabenId,Loesung),
		   [],
		   ausw([],[],_),
		   ausw(PosMap,VarMap,ML),
		   0,100,Fehler),
	session_setValue(posMap,PosMap),
	session_setValue(varMap,VarMap),
	session_setValue(musterL,ML),
	logging(auswertung,'<auswertung>Output: Fehler=~w, Musterloesung=~w',[Fehler,ML]).


auswertung0(AufgabenId,Loesung)-->
	{musterloesung(AufgabenId,ML0),
	 normalform:normalizeBody(ML0,Musterloesung,[])},
	match_bags(Loesung,Musterloesung,PosMap,zuordnen,20),
	setf(ids_state^state2-ausw^posMap,PosMap),
	setf(ids_state^state2-ausw^musterL,Musterloesung).

auswertung_pred(Loesung,AufgabenId,Fehler):-
	ids_search(auswertung_pred0(AufgabenId,Loesung),
		   [],
		   ausw([],[],_),
		   ausw(PosMap,_,ML),
		   0,100,Fehler),
	session_setValue(posMap,PosMap),
	session_setValue(musterL,ML).

auswertung_pred0(AufgabenId,Loesung)-->
	{musterloesung(AufgabenId,ML0),
	 normalform:normalizeBody(ML0,Musterloesung,[])},
	match_bags(Loesung,Musterloesung,PosMap,zuordnen_pred,20),
	state2(_,ausw(PosMap)).

% this part is used for analyzing two predicates.
analyse_predicates(StudentPred,SamplePred,Fehler):-
	logging(auswertung,'<auswertung>Input: StudentPred=~w, SamplePred=~w',[StudentPred,SamplePred]),
	ids_search(match_predicates2(StudentPred,SamplePred),
		   [],
		   ausw([],[],_),
		   ausw(PosMap,VarMap,_),
		   0,100,Fehler),
	session_setValue(posMap,PosMap),
	session_setValue(varMap,VarMap),
	logging(auswertung,'<auswertung>Output: Fehler=~w',[Fehler])
	.

%match_predicates(Pred1,Pred2)-->
%	with_position(1,match_bags(Pred1,Pred2,match_clauses,15)).

match_predicates2(StudentPred,SamplePred)-->
	%with_position(1,match_bags(StudentPred,SamplePred,PosMap,match_clauses,20)),
	match_bags(StudentPred,SamplePred,PosMap,match_clauses,20),
	setf(ids_state^state2-ausw^posMap,PosMap),
	setf(ids_state^state2-ausw^musterL,SamplePred).
        
match_clauses(Head1-Body1,Head2-Body2)-->
%	setf(ausw^posMap,[]),
	setf(ids_state^state2-ausw^varMap,[]),
	with_position(0,zuordnen(Head1,Head2)),
	match_bags(Body1,Body2,zuordnen,10).

% end of: this part is used for analyzing two predicates


%=zuordnen_pred(+Term,+Term2) (*)
%
% tries to match two subgoals
% (one from the sample solution and one from the
%  student solution)
% Because we know that there are type errors, we just look at the
% name of the functor.
zuordnen_pred(Term,Term2)-->
	{functor(Term,Functor,_),
	 functor(Term2,Functor,_)}.
zuordnen_pred(Term,Term2)-->
	{Term=..[Op1,_,_],
	 Term2=..[Op2,_,_],
	 once(infer_types:compare_op(Op1);
	      infer_types:faulty_compare_op(Op1,_)),
	 once(infer_types:compare_op(Op2);
	      infer_types:faulty_compare_op(Op2,_))}.


%=zuordnen(+Term,+Term2) (*)
%
% tries to match two subgoals
% (one from the sample solution and one from the
%  student solution)
zuordnen(Term,Term2)-->
	{functor(Term,Functor,Arity),
	 functor(Term2,Functor,Arity),
	 logging(auswertung,'<zuordnen1>Term=~w und Term2=~w have the same functor=~w',[Term,Term2,Functor]),
	 not(infer_types:compare_op(Functor)),
	 Term=..[_|Args],
	 Term2=..[_|Args2]},
	match_lists_(Args,Args2,
		     zuordnen_arg,
		     unmatchable,3),
	{logging(auswertung,'<zuordnen1>Term und Term2 are matched',[])}.
% match two lists.
zuordnen([X|Xs],[Y|Ys])-->
	{logging(auswertung,'<zuordnen2>Term1=[~w|~w] und Term2=[~w|~w]',[X,Xs,Y,Ys])},
	zuordnen_arg(X,Y),
	zuordnen_arg(Xs,Ys),
	{logging(auswertung,'<zuordnen2>Term1 and Term2 are matched',[])}.
	

zuordnen(Ineq1,Ineq2)-->
	{xform_ineq(Ineq1,PV1,NV1,N1),
	 xform_ineq(Ineq2,PV2,NV2,N2)},
	match_ineq(PV1,NV1,N1,PV2,NV2,N2).

zuordnen(A is TermA, B is TermB)-->
	zuordnen_arg(A,B),
	{xform_term(TermA,PV1,NV1,N1),
	 xform_term(TermB,PV2,NV2,N2)},
	match_bags(PV1,PV2,zuordnen_arg,2),
	match_bags(NV1,NV2,zuordnen_arg,2),
	expect1(N1=:=N2,ineq_offset,2).

zuordnen(A1@<A2,B1@>B2)-->
	zuordnen_arg(A1,B2),
	zuordnen_arg(A2,B1).
zuordnen(A1@>A2,B1@<B2)-->
	zuordnen_arg(A1,B2),
	zuordnen_arg(A2,B1).
	
match_ineq(PV1,NV1,N1,
	   PV2,NV2,N2)-->
	match_bags(PV1,PV2,zuordnen_arg,2),
	match_bags(NV1,NV2,zuordnen_arg,2),
	expect1(N1=:=N2,ineq_offset,2).

match_ineq(PV1,NV1,N1,
	   NV2,PV2,N2)-->
	error(inverted_ineq,1),
	match_bags(PV1,PV2,zuordnen_arg,2),
	match_bags(NV1,NV2,zuordnen_arg,2),
	expect1(N1=:= -N2,ineq_offset,2).

%=xform_ineq(Goal,PosVar,NegVar,N)
% reduces a simple inequality with (hopefully) only integers
% e.g. A+1=<4-B+C becomes
% PosVar=[C], NegVar=[B,A], N=3
xform_ineq(A<B,PV,NV,N):-
	xform_term(B-A-1,PV,NV,N).
xform_ineq(A>B,PV,NV,N):-
	xform_term(A-B-1,PV,NV,N).
xform_ineq(A=<B,PV,NV,N):-
	xform_term(B-A,PV,NV,N).
xform_ineq(A>=B,PV,NV,N):-
	xform_term(A-B,PV,NV,N).

xform_term(A+B,PV,NV,N):-
	xform_term(A,PV1,NV1,N1),
	xform_term(B,PV2,NV2,N2),
	append(PV1,PV2,PV),
	append(NV1,NV2,NV),
	N is N1+N2.
xform_term(A-B,PV,NV,N):-
	xform_term(A,PV1,NV1,N1),
	xform_term(B,NV2,PV2,N2),
	append(PV1,PV2,PV),
	append(NV1,NV2,NV),
	N is N1-N2.
xform_term(V,[V],[],0):-
	memberchk(V,['<var>'(_),'<a-var>'(_)]).
xform_term(N,[],[],N):-
	integer(N).

zuordnen_arg('<a-var>'(_),'<a-var>'(_))-->[].
zuordnen_arg('<var>'(Name1),'<var>'(Name2))-->
	modify(ids_state^state2-ausw^varMap,Vars0,Vars1),
	{add_to_bimap(Name1,Name2,Vars0,Vars1)}.
zuordnen_arg(['<var>'(Head1)|'<var>'(Tail1)],['<var>'(Head2)|'<var>'(Tail2)])-->
	zuordnen_arg('<var>'(Head1),'<var>'(Head2)),
	zuordnen_arg('<var>'(Tail1),'<var>'(Tail2)).
zuordnen_arg(X,Y)-->
	{atomic(X),atomic(Y)},
	expect1(X=Y,misspelled(X,Y),2).


%=occurs0(Varname,Term)
% succeeds if the variable named Varname
% occurs in Term and Term is not an arithmetic operation
% (i.e. Term is a call to a DB predicate that grounds Var).
% <b>in general, the assumption that all other predicates are DB predicates
% is a bit too optimistic, as in
% d(A,B,C)@<d(2003,12,13),A<2000.</b>
% Example Term= X is Y+4 => Varname=X
occurs0(Varname,Term):-
	%logging(auswertung, '<OCCURS0> Term=~p',[Term]),
	Term=..[Fun|Args],
	not(infer_types:compare_op(Fun)),
	Fun\=is,
	member(Arg,Args),
	%logging(auswertung, '<OCCURS0> Arg=~p; Args=~p',[Arg,Args]),
	occurs(Varname,Arg),
	logging(auswertung, '<OCCURS0> There is a variable Varname=~p as operand of operator ~p',[Varname,Fun]).
occurs0(Varname,'<var>'(Varname) is _).

%=occurs1(Varname,Term)
% succeeds if the variable named Varname
% occurs in Term in a position where it has to be
% grounded (left side of 'is', in a comparision etc.)
% Example: Term=N>M => Varname=N;Varname=M 

occurs1(Varname,Term):-
	Term=..[Fun,Arg1,Arg2],
	infer_types:compare_op(Fun),
	logging(auswertung, '<OCCURS1> operator is ~p',[Fun]),
	(
	occurs(Varname,Arg1);
	occurs(Varname,Arg2)
	),
	logging(auswertung, '<OCCURS1> There is a variable Varname=~p as operand of operator ~p',[Varname,Fun]).
	
occurs1(Varname,Term):-
	
	Term=..[Fun,Arg1,Arg2],
	instantiation_sensitive(Fun),
	logging(auswertung, '<OCCURS1> operator is ~p',[Fun]),
	(
	occurs(Varname,Arg1);
	occurs(Varname,Arg2)
	),
	logging(auswertung, '<OCCURS1> There is a variable Varname=~p as operand of operator ~p',[Varname,Fun]).

occurs1(Varname,_ is That):-
	occurs(Varname,That),
	logging(auswertung, '<OCCURS1> There is a variable Varname=~p as operand of operator IS',[Varname]).		

% check if a term contains a equality operation
%= occursEqual(-Varname, +Term)

occursEqual(Varname, Term):-
	Term=..[Functor,Arg1,Arg2],
	Functor=='=',
	logging(auswertung, '<OCCURSEQUAL> Functor is ~p, Arg1=~p, Arg2=~p',[Functor,Arg1,Arg2]),
	(
	nonvar(Arg1)->occurs(Varname,Arg1);
	nonvar(Arg2)->occurs(Varname,Arg2)
	),
	logging(auswertung, '<OCCURSEQUAL> There is a variable Varname=~p as operand of operator ~p',[Varname,Functor]).

%=occurs(Varname,Term)
% succeeds if the variable named Varname
% occurs in Term or a subterm of Term
occurs(Varname,'<var>'(Varname)).
occurs(Varname,Term):-
	Term=..[_|Args],
	member(Arg,Args),
	occurs(Varname,Arg).



%=checkInst(Terms,Errs)
% Example: X>1, ausleihe(_,_,_,X,_) will produce errors.
% Y=1,ausleihe(_,_,_,X,_), Y=2 

checkInst(Terms,Errs):-
	checkInst(1,Terms,[],Errs).
checkInst(_,[],_,[]).
checkInst(N,[Term|Terms],Before,Errs):-
	logging(auswertung, '<CHECKINST> Term=~p and Terms=~p',[Term,Terms]),
	logging(auswertung, '<CHECKINST> Before=~p',[Before]),
	% N stands for N.th subgoal.
	findall(error([N],uninstantiated(Varname)), 
		 (occurs1(Varname,Term),    % variable as operand of >,< exist here, but does not exist in the last subgoals
		  not((member(Term2,Before),occurs0(Varname,Term2)))),
		NewErrs), 
	logging(auswertung, '<CHECKINST> Errors after check with occurs1 and occurs0=~p',[NewErrs]),
		
	% check monotonicity
	findall(error([N],already_instantiated(Varname)), 
		 (occursEqual(Varname,Term),    % an instantiated variable is found
		  ((member(Term2,Before),occursEqual(Varname,Term2)))),
		NewErrs1), 
	% then concatenate the error lists after checking instantiatiton.
	append(NewErrs, NewErrs1,ErrsTemp),
	append(ErrsTemp,Errs2,Errs),
	
	succ(N,N2),
	checkInst(N2,Terms,[Term|Before],Errs2).
	 
	

