% Author: Yannick Versley, Nguyen-Thinh Le
% Modified by: $Id: diagnose_db.pl,v 1.2 2004/03/12 16:20:12 nthinh Exp $



:- module(diagnose_db,[diagnose_db/4]).
:- use_module(logging).

%=diagnose_db(+Id,+Anfrage,-ErrorList, -CorrectedTerm)
% gathers the diagnosis information from syntax checker, type checker, instantiation checker.
% Id is the id of the exercise
% Anfrage is the student solution
% Errorlist contains the errors the system found in the student solution
% CorrectedTerm is the possible correction the system provides.

diagnose_db(Id,Anfrage,Errs, CorrectedTerm):-
	% Syntax checking
	logging(diagnose_db,'<diagnose_db>Anfrage is a string that is bound to a term to be checked: Anfrage=~w',[Anfrage]),
	atom_codes(Anfrage,AnfrageCharList),
	check_query_syntax(AnfrageCharList,TermL,SyntaxErrs)->
	logging(diagnose_db,'<diagnose_db>Term after checked: TermOut=~w and Error list=~w',
	      [TermL,SyntaxErrs]),
	logging(diagnose_db,'<diagnose_db>Term after normailized: TermL=~w',[TermL]),
	% returns this normalized and corrected term back

	CorrectedTerm=TermL,
	(       SyntaxErrs==[] ->
	(      % Type checking
	typify_query(TermL,Errors, VarTypes),
	logging(diagnose_db,'<diagnose_db> Errors after typify check: Errors=~w', [Errors]),
        logging(diagnose_db,'<diagnose_db>before calling checkInst TermL2=~p', [TermL2]),
	logging(diagnose_db,'<diagnose_db>ErrsCI=~p', [ErrsCI]),
	session_setValue(vartypes,VarTypes),     
	logging(diagnose_db,'<diagnose_db><DIAGNOSE>vartypes=~p', [VarTypes]),
	% Instantiation checker
	checkInst(TermL,ErrsCI),

	% Unification equations simplifier
	simplifyUni(Uni,TermL,TermL2),
	logging(diagnose_db,'<diagnose_db>Uni=~p', [Uni]),
        logging(diagnose_db,'<diagnose_db>TermL=~p', [TermL]),
        logging(diagnose_db,'<diagnose_db>TermL2=~p', [TermL2]),
	maplist(uni2warn,Uni,UniErr),

	(Errors=[]->
	% Full match between the solutions
	 once(auswertung(TermL2,Id,Errs2);Errs2=[error([],auswertung_failed)]),
	 flatten([SyntaxErrs,UniErr,ErrsCI,Errs2],Errs);
	% Only a partial match between the solutions
	once(auswertung_pred(TermL2,Id,Errs2);
	     Errs2=[error([],auswertung_failed)]),
        flatten([UniErr,ErrsCI,error([],typify_errors),Errs2,Errors],Errs))
	%;
	
	);
	(   logging(diagnose_db,'<diagnose_db>Only syntax errors: ~w',[SyntaxErrs]),
	Errs=SyntaxErrs
	));
	% else, syntax checker fails
	Errs=[error([],syntax_analysis_failed)],
	CorrectedTerm=erronous_term
	.
