:- module(diagnose_rec,[diagnose_rec/4]).
:- use_module(logging).

diagnose_rec(Id,Program,Errs,Hint):-!,
	logging(diagnose_rec,'<diagnose_rec>Input value Id=~w',[Id]),
        logging(diagnose_rec,'<diagnose_rec>Input value Program=~w',[Program]),
	(   Id='nix'->(Errs=[error([],no_exercise_chosen)], Hint="no hint");
	              assume(ground(Id)),
                      assume(ground(Program)),
	              atom_codes(Program,ProgramC),
	              logging(diagnose_rec,'<diagnose_rec>syntax checking begins with program=~w',[ProgramC]),
	              check_program_syntax2(ProgramC,Clauses,ErrsSyn)->
	              (	 logging(diagnose_rec,'<diagnose_rec>syntax errors=~w',[ErrsSyn]),
	                 %logging(diagnose_rec,'<diagnose_rec>Clauses=~w',[Clauses]),
                         append(ErrsSyn,Errs1,Errs),
	                 after_syntax(Id,Clauses,Errs1,Hint)
		      );
	              Errs=[error([],syntax_check_failed)]
	).

after_syntax(Id,Clauses,Errs,Hint):-
	logging(diagnose_rec,'<after_syntax>Input value Id=~w',[Id]),
        logging(diagnose_rec,'<after_syntax>Input value Clauses=~w',[Clauses]),
	typify_pred(Clauses,PredTypes,ErrsTyp),
	logging(diagnose_rec,'<after_syntax>PredTypes=~w',[PredTypes]),
	logging(diagnose_rec,'<after_syntax>ErrsTyp=~w',[ErrsTyp]),
	!,(
	  session_setValue(pred_types,PredTypes),
	  ErrsTyp=[]->
	    logging(diagnose_rec,'No type errors.',[]),
    	    after_typify(Id,Clauses,Errs,Hint);
	  ErrsTyp=Errs).
%	session_getValue(posMap,PosMap).
%	logging(diagnose_rec,'<after_syntax>PosMap=~w',[PosMap]).
after_syntax(_,_,[error([],typify_failed)]).

after_typify(Id,StudentClauses,Errs,Hint):-
	logging(diagnose_rec,'<after_typify>Input value Id=~w, Clauses=~w',[Id,StudentClauses]),
        musterloesung(Id,SampleClauses),!,
	analyse_predicates(StudentClauses,SampleClauses,StructureErrors),
	logging(diagnose_rec,'<after_typify>StructureErrors=~w',[StructureErrors]),
	session_getValue(posMap,PosMap),
	logging(diagnose_rec,'<after_typify>PosMap=~w',[PosMap]),
	% PosMap=[0=1,1=2,2=0] shows the possible matching between the student and sample clauses
	% Student clauses are not needed to be sorted
	sortClauses(PosMap,SampleClauses,SampleClauses,SortedSampleClauses),
	logging(diagnose_rec,'<after_typify>SortedSampleClauses=~w',[SortedSampleClauses]),
	pattern(Id,Schema),
	logging(diagnose_rec,'<after_typify>Schema=~w',[Schema]),
	analyse_idiom(Schema,StudentClauses,SortedSampleClauses,1,ErrsIdiom),
	logging(diagnose_rec,'<after_typify>IdiomErrs=~w',[ErrsIdiom]),
	(   nonvar(ErrsIdiom)->
	    (   (   StructureErrors\=[];ErrsIdiom\=[])->patterndescription(Id,Hint);
	    Hint='your solution was successful.'
	    ),
	    append(StructureErrors,ErrsIdiom,Errs);
	    Errs=StructureErrors
	),
	logging(diagnose_rec,'<after_typify>Output Errs=~w',[Errs]),
	logging(diagnose_rec,'<after_typify>Output Hint=~w',[Hint]).
after_typify(_,_,[error([],no_exercise_chosen)]).

sortClauses([_StudentClausePos=SampleClausePos|PosRest],InClauses, Accu,[SampleClause|OutClauses]):-
	logging(diagnose_rec,'<sortClauses begins>SampleClausePos=~w, PosRest=~w, Accu=~w, InClauses=~w',[SampleClausePos,PosRest,InClauses,Accu]),
	nth0(SampleClausePos,InClauses,SampleClause),
	logging(diagnose_rec,'<sortClauses>SampleClause=~w',[SampleClause]),
	select(SampleClause,Accu,Accu1),
	logging(diagnose_rec,'<sortClauses>Accu1=~w',[Accu1]),
	sortClauses(PosRest,InClauses,Accu1,OutClauses).
sortClauses([],_,Accu,Accu).

	
	
	



