% Author: Yannick Verley
% Modified: $Id: domain.pl,v 1.31 2004/04/05 11:39:42 nthinh Exp $

% % This module provides a means to load the descriptions of database query and recursion exercises.




:- [auswertung].
:- use_module(infer_types).
:- use_module(struct).
:- use_module(syntaxChecker).
:- use_module(logging).

:- dynamic domain/3.
:- dynamic rekursion/2.
:- dynamic aufgabe/2.
:- dynamic musterloesung/2.
:- dynamic testfall/3.
:- dynamic feedback/3.
:- dynamic fehlertyp/3.
:- dynamic current_domain/1.
:- dynamic pattern/2.
:- dynamic patterndescription/2.

%%
%% loading of the domain description
%%

load_domain_xml(Filename):-
	load_xml_file(Filename,Tree),
	consumeTree(Tree).

%=consumeTree(+XMLTags)
% XMLTags: a list
% parsing of the domain description (top level)
% element() is a functor, see docu of sgml library.

consumeTree([]).
consumeTree([element(domain,[name=Dom,
			     db=DBFile,types=Typefile],
		     Exercises)|Rest]):-
	set_domain(Dom),
	retractall(domain(Dom,_,_)),
	assert(domain(Dom,DBFile,'es gibt noch keine Beschreibung')),
	consult(Typefile),
	consult(DBFile),
	retractall(aufgabe(Dom-_,_)),
	retractall(musterloesung(Dom-_,_)),
	retractall(testfall(Dom-_,_,_)),
	retractall(xml_aufgabe(Dom-_,_)),
	consumeExercises(Dom,0,Exercises),
	consumeTree(Rest).
consumeTree([element(rekursion,Attrs,
		     Exercises)|Rest]):-
	memberchk(name=Dom,Attrs),
	set_domain(Dom),
	retractall(rekursion(Dom,_)),
	assert(rekursion(Dom,'es gibt noch keine Beschreibung')),
	(memberchk(types=Typefile,Attrs)->
	consult(Typefile);
	true),
	retractall(aufgabe(Dom-_,_)),
	retractall(musterloesung(Dom-_,_)),
	retractall(testfall(Dom,_,_)),
	retractall(xml_aufgabe(Dom-_,_)),
	consumeExercises2(Dom,0,Exercises),
	consumeTree(Rest).

%= consumeExercises2(+Domain,+ExerciseNumber,+XMLTags)
% this consumes the exercises of recursion part
% Domain: the domain of the exercise (e.x rekursion )
% ExerciseNumber: the last exercise number (the 1 from bib-1). Next
%                 exercise number is ExerciseNumber + 1.
% parses the domain description and the exercises

consumeExercises2(_,_,[]).
consumeExercises2(Dom,N,[Whitespace|Rest]):-
	atom(Whitespace),!,
	consumeExercises2(Dom,N,Rest).
consumeExercises2(Dom,N,[element(description,[],[Descr])|Rest]):-
	write(Descr),
	retract(rekursion(Dom,_)),
	assert(rekursion(Dom,Descr)),
	consumeExercises2(Dom,N,Rest).

consumeExercises2(Dom,N,[element(aufgabe,[],Aufgabe)|Rest]):-
	succ(N,N1),
	consumeAufgabe2(Dom,N1,Aufgabe),
	consumeExercises2(Dom,N1,Rest).

%= consumeExercises(+Domain,+ExerciseNumber,+XMLTags)
% this consumes the exercises of database query part.
% Domain: the domain of the exercise (e.x bib )
% ExerciseNumber: the last exercise number (the 1 from bib-1). Next
%                 exercise number is ExerciseNumber + 1.
% parses the domain description and the exercises

consumeExercises(_,_,[]).
consumeExercises(Dom,N,[Whitespace|Rest]):-
	atom(Whitespace),!,
	consumeExercises(Dom,N,Rest).
consumeExercises(Dom,N,[element(description,[],[Descr])|Rest]):-
	write(Descr),
	retract(domain(Dom,File,_)),
	assert(domain(Dom,File,Descr)),
	consumeExercises(Dom,N,Rest).

consumeExercises(Dom,N,[element(aufgabe,[],Aufgabe)|Rest]):-
	succ(N,N1),
	consumeAufgabe(Dom,N1,Aufgabe),
	consumeExercises(Dom,N1,Rest).

%=consumeAufgabe(+Domain,+ExerciseNumber,+XMLTags)
% Domain, ExerciseNumber: the two parts that identify the
%                         exercise (e.g. bib-1)
% parses the contents of an <aufgabe> tag
% including the sample solution, the
% exercise description and the test cases
consumeAufgabe(_,_,[]).
consumeAufgabe(Dom,N,[Whitespace|Rest]):-
	atom(Whitespace),!,
	consumeAufgabe(Dom,N,Rest).
consumeAufgabe(Dom,N,
	       [element(text,[],Text)|Rest]):-!,
	generate_text(Text,
		      gt([],[],Plain0),
		      gt(_,_,[])),
	concat_atom(Plain0,Plain1),
	atom_chars(Plain1,Plain2),
	maplist(replaceWhite,Plain2,Plain3),
	atom_chars(Plain,Plain3),
	assert(aufgabe(Dom-N,Plain)),
	assert(xml_aufgabe(Dom-N,Text)),
	consumeAufgabe(Dom,N,Rest).
consumeAufgabe(Dom,N,
	       [element(loesung,[],[Loesung])|Rest]):-!,
	format('Musterlsung: ~w\n',[Loesung]),
	consumeAtom1(Loesung,LoesungC),
	assert(musterloesung(Dom-N,LoesungC)),
	consumeAufgabe(Dom,N,Rest).
consumeAufgabe(Dom,N,
	       [element(testfall,[],_Testfall)|Rest]):-!,
	consumeAufgabe(Dom,N,Rest).
%consumeAufgabe(Dom,N,
%       [element(feedback,[],[_Feedback])|Rest]):-!,
%	consumeAufgabe(Dom,N,Rest).
consumeAufgabe(Dom,N,
	       [Element|_]):-!,
	print_message(error,format('consumeAufgabe(~w-~w): unknown Element ~w',
				   [Dom,N,Element])).

%=consumeAufgabe2(+Domain,+ExerciseNumber,+XMLTags)
% Domain, ExerciseNumber: the two parts that identify the
%                         exercise (e.g. bib-1)
% parses the contents of an <aufgabe> tag
% including the sample solution, the
% exercise description and the test cases
consumeAufgabe2(_,_,[]).
consumeAufgabe2(Dom,N,[Whitespace|Rest]):-
	atom(Whitespace),!,
	consumeAufgabe2(Dom,N,Rest).
consumeAufgabe2(Dom,N,
		[element(text,[],Text)|Rest]):-!,
	assert(xml_aufgabe(Dom-N,Text)),
	consumeAufgabe2(Dom,N,Rest).
consumeAufgabe2(Dom,N,
		[element(titel,[],[Text])|Rest]):-!,
	format('Aufgabe ~w-~w: ~w\n',[Dom,N,Text]),
	assert(aufgabe(Dom-N,Text)),
	consumeAufgabe2(Dom,N,Rest).
consumeAufgabe2(Dom,N,
	       [element(loesung,[],Loesung)|Rest]):-!,
	consumeSolution(Dom,N,Loesung),
	consumeAufgabe2(Dom,N,Rest).
consumeAufgabe2(Dom,N,
	       [element(testfall,[],[TestfallA])|Rest]):-!,
	consumeAtom1(TestfallA,Testfall),
	assert(testfall(Dom,N,Testfall)),
	consumeAufgabe2(Dom,N,Rest).
consumeAufgabe2(Dom,N,
	       [Element|_]):-!,
	print_message(error,
		      format('consumeAufgabe2(~w-~w): unknown Element ~w',
			     [Dom,N,Element])).

consumeSolution(_,_,[]).
consumeSolution(Dom,N,[Whitespace|Rest]):-
	atom(Whitespace),!,
	consumeSolution(Dom,N,Rest).

consumeSolution(Dom,N,[element(pattern,[],[Pattern])|Rest]):-!,
	format('Pattern of sample solution: ~w',[Pattern]),
	%retract(pattern(Dom-N,_)),
	assert(pattern(Dom-N,Pattern)),
	consumeSolution(Dom,N,Rest).
consumeSolution(Dom,N,[element(patterndescription,[],[PatternDescription])|Rest]):-!,
	format('Pattern description of sample solution: ~w',[PatternDescription]),
	%retract(patterndescription(Dom-N,_)),
	assert(patterndescription(Dom-N,PatternDescription)),
	consumeSolution(Dom,N,Rest).

consumeSolution(Dom,N,[element(samplesolution,[],[Loesung])|Rest]):-!,
	format('Musterlsung: ~w',[Loesung]),
	consumeAtom(Loesung,LoesungC),
	maplist(normalize,LoesungC,LoesungN),
	assert(musterloesung(Dom-N,LoesungN)),
	consumeSolution(Dom,N,Rest).


replaceWhite('\n',' '):-!.
replaceWhite(Char,Char).


%%
%% gathering of the errror terms from various components
%%

%=uni2warn(+Equation,-Warning)
% simply wraps the Var=SomeTerm terms as given by simplifyUni into
% an error term
uni2warn(A,warning([],substituted(A))).



