% Author: Yannick Versley, Nguyen-Thinh Le
% Modified by: $Id: feedbackHandler.pl,v 1.6 2004/03/29 07:41:27 nthinh Exp $


:-module(feedbackHandler,[feedback/5,return_the_feedbacklist/2, generate_text/3, locate_error_in_term/2, write_string_list/1]).
:- use_module(listUtil).
:- use_module(logging).


%= public feedback(-Priority, +ErrorTerm, -Where, -What, -How)
% Priority of the feedback for this error, 
% ErrorTerm has the structure error([ErrorPosition], ErrorDescription)
% Where is the error
% What is the error
% How to handle this error 

% error position 
% error([1, 1, 2], unmatchable(1, <a-var>(_311), 1, <var>(Xs))) means
% in clause 2 (last element of the list), body (the second element),
% first argument is false. In the student solution it's a anonymous
% variable, in the sample solution its a variable.

% error([1,0,2], unmatchable(1, [<var>(X)|<var>(Xs)],1,[<var>(X)|<var>(Xs)])) 
% in clause 2, in the head (the second element of the position list),
% the first argument is false.

% Errors which can be identified and are solvable have the higher priority.

%%%%%% Errors for structure check, indicators: 1-20 %%%%%%%%%%%%

feedback(1, error(_,missing('(')),'(','In deiner Eingabe fehlt dieses Zeichen.','Das verursacht Syntaxfehler.').
feedback(2, error(_,missing(')')),')','In deiner Eingabe fehlt dieses Zeichen.','Das verursacht Syntaxfehler.').
feedback(3, error(_,missing('[')),'[','In deiner Eingabe fehlt dieses Zeichen.','Das verursacht Syntaxfehler.').
feedback(4, error(_,missing(']')),']','In deiner Eingabe fehlt dieses Zeichen.','Das verursacht Syntaxfehler.').
feedback(5, error(_,missing(',')),',','In deiner Eingabe fehlt dieses Zeichen.','Das verursacht Syntaxfehler.').

feedback(10, error(_,misplaced_list_element(_,Konst,_,_)),Konst, 'Diese Konstante befindet sich an der falschen Argumentstelle.','Bitte korrigiere es.').

feedback(11, error(_,misplaced_list_element(_,'<a-var>'(_),_,_)),_,'die Argumentreihenfolge stimmt nicht', 'Bitte prfe es nach.'):-!.

feedback(12, error(_,misplaced_list_element(_,'<var>'(_),_,_)),_,'die Argumentreihenfolge stimmt nicht', 'Bitte prfe es nach.'):-!.

feedback(13, error([_,No],superfluous_list_elements),No, 'Dieses Teilziel erwartet weniger Argumente','Prfe die Stelligkeit des Teilziels nach.').

feedback(14, error([_,SubgoalPos],forgotten_list_element(_,_)),[SubgoalPos,'.Teilziel'],'Dieses Teilziel erwartet mehr Argumente','Prfe die Stelligkeit des Teilziels nach.').

feedback(15, error(_,misspelled_predicate(Misspelled,Corrected)),Misspelled,'Dieses Prdikat gibt es nicht.', ['Meintest du <b>',Corrected,'</b>?']).

% zu wenig Teilziele
feedback(16,error([SubgoalPos],unmatched_set2(_)),[term(SubgoalPos),'. Teilziel'],'Dies reicht noch nicht.','Bitte prfen, wir mssen mehr Teilziele haben.').

feedback(17,error([_],unmatched_set1(Term)),Term, 'Brauchst Du diesen Term?', 'Es scheint berflssig zu sein.' ).

%feedback(15,error(_,unknown_predicate(Pred)),  Pred,'Das Praedikat kenne ich nicht.', 'Bitte korrigiere den Prdikatname.').

feedback(18,error(_,wrong_head(Head,_)),Head, 'Irgendetwas stimmt mit diesem Klauselkopf nicht.','Entweder Du hast "." und "," verwechselt, oder Du hast ein ":-" falsch geschrieben').

%%%%%%%%% Semantic errors, indicators: 21-80 %%%%%%%%%%%

feedback(21,error(_,atom_as_var(Atom)),Atom, 'Soll das eine Variable sein?','dann muss der erste Buchstabe gross geschrieben werden.').
% ???
feedback(23, error(_,not_inside_comp(Op)),Op,'diesen Vergleich musst Du in ein eigenes Teilziel tun.',' Bitte korrigiere es.').


% ???
feedback(25, error(_,not_inline_comp(Op)),Op,'diesen Vergleich musst Du in ein eigenes Teilziel tun.',' Bitte korrigiere es.').


feedback(27, error(_,inline_arithmetic(Term)),Term,'Diesen arithmetischen Ausdruck musst Du in ein eigenes Teilzeil tun.', 'Bitte korrigiere es.').

feedback(30,error(_,conflicting_type(Varname,_Typ)),Varname,'Diese Variable kann nicht innerhalb eines Ziels gleichzeitig mehrere Typen haben', ['Siehe typischen Typkonfliktsfehler ', ConflictTypeList ]):-
	session_getValue(vartypes, VarTypes),
	findall(Varname=ConflictType, member(Varname=ConflictType,VarTypes), ConflictTypeList).
	% VarTypes has something like this: vartypes=['Bla'=signatur, 'Bla'=datum, 'Sig'=signatur]
	

% ???
feedback(32, error(_,must_be_number(Term)),Term, 'Hast Du das als ein Zahl gemeint?', 'An dieser Stelle muss eine Zahl sein.').

% ???
feedback(34, error(_,no_arithmetic_function(Term)),Term,' Dies ist kein arithmetischer Ausdruck.', 'Bitte prfe nach.').

% more info for how
feedback(39, error(_,use_X_not_Y(Op1,Op2)),Op2, 'Diese Operation ist vermutlich nicht richtig', ['Du meintest bestimmt <b>',Op1,'</b>']).

feedback(45, error(_,uninstantiated(Var)),Var, 'Diese Variable ist noch nicht instanziiert.', 'Diese Variable sollte instanziiert sein, bevor sie fr eine arithmetische Operation verwendet wird').

feedback(46, error(_,already_instantiated(Var)),Var,'Diese Variable ist schon instantiert.','Eine Variable kann nur ein Mal mit einem Wert gebunden werden.').

% ???
feedback(48, error(_,ineq_offset),_,'','Schau dir die Ungleichung noch einmal ganz genau an.').


coreference_error(Errs,
		  40-['Beim ',N1,'. und beim ',N2,'. Teilziel sollte ',
		      term('<var>'(Varname)),
		      ' dasselbe beinhalten (Koreferenz).']):-
	member(error([_,N1],unmatchable(_,'<a-var>'(_),_,'<var>'(Varname))),
	       Errs),
	member(error([_,N2],unmatchable(_,'<a-var>'(_),_,'<var>'(Varname))),
	       Errs),
	N1<N2.

%%% unmatchable errors %%%%%%%%%%%%%%%%%%


% Student: Constant; Sample: Variable 
feedback(50, error(_,unmatchable(_,Const,_,'<var>'(_))), Const, 'War diese Information in der Aufgabe gegeben?', 'An dieser Stelle sollte es eine Variable sein.').

% Student's solution: Variable; Sample: constant
feedback(54, error(_,unmatchable(_,'<var>'(Var),_,Const)),Var , 'An dieser Stelle sollte es eine Konstante sein.', 'Finde diese Angabe in der Aufgabestellung.'):- (atomic(Const); Const=.._),!.

% Student's solution: anonymous Variable; Sample constant
feedback(58, error(_,unmatchable(_,'<a-var>'(AVar),_,Const)),AVar , 'An dieser Stelle sollte es eine Konstante sein.', 'Finde diese Angabe in der Aufgabestellung.'):- (atomic(Const); Const=.._),!.

% Student's solution: constant; Sample: anonymous 
feedback(60, error(_,unmatchable(_,Const,_,'<a-var>'(_))), Const, 'Diese Information brauchst Du nicht, oder?', 'Du kannst an dieser Stelle eine anonyme Variable haben.').

% Student's solution: variable; Sample: anonymous 
feedback(62, error(_,unmatchable(_,'<var>'(Var),_,'<a-var>'(_))), Var, 'Diese Information brauchst Du nicht, oder?', 'Du kannst an dieser Stelle eine anonyme Variable haben.').

% Student's solution: anonymous Variable, Sample solution: non-anonymous
feedback(64,error(_,unmatchable(_,'<a-var>'(AVar),_,'<var>'(_))), AVar, 'Du hast eine anonyme Variable angegeben', 'An dieser Stelle sollte eine non-anonyme Variable sein.').

% coferefence is required
feedback(68, error(_,unmatchable(_,'<var>'(Varname1),
				  _,'<var>'(MVarname))),[Varname1, ',',Varname2], 'Koreferenz existiert', 'Diese Variablen sollten miteinander koreferieren.'):-
	session_getValue(varMap,VarMap),
	member(Varname2=MVarname,VarMap).

% coreference is not required ???

%%% substituted errors %%%%%%%%%%%%%%%%%%%%%%%%%%
feedback_term(warning(_,substituted('<var>'(X)=K)),
	      80-
	      ['Die Variable <b>',
	       term('<var>'(X)),
	       '</b> ist berflssig. Es kann durch die Konstante <b>',
	       term(K),
	       '</b> ersetzt werden']).
feedback(80, error(_,substituted('<var>'(X)=K)),X,' Diese Variable ist berflssig.',['Es kann durch die Konstante <b>',
	       term(K),'</b> ersetzt werden']).
feedback_term(warning(_,substituted(K='<var>'(X))),
	      80-
	      ['Die Variable <b>',
	       term('<var>'(X)),
	       '</b> ist berflssig. Es kann durch die Konstante <b>',
	       term(K),
	       ' </b> ersetzt werden']).
feedback(81, error(_,substituted(K='<var>'(X))),X,'Die Variable ist berflssig.', ['Es kann durch die Konstante <b>',
	       term(K),' </b> ersetzt werden']).
%%%%%%%%%%%% errors 90 %%%%%%%%%%%%%%%%%
% errors without any idea


feedback(98,error(_,unknown_predicate(erronous_term)),'Fehler kann nicht identifiziert werden.','Ist da eine Katze uber die Tastatur gelaufen?','Bitte konzentriere dich auf die Aufgabestellung'):-!.

feedback(100,error(_,typify_errors),'Fehler kann nicht identifiziert werden.','Deine Anfrage passt nicht zur Datenbank.','Bitte prfe die Prdikate nach.').


% -------------------------------------------------------------------
% public write_string_list(+List)
% This predicate is used for writing the feedback terms
% List is a list of strings that can contain term(Something)
write_string_list(List):-
	is_list(List)->
	        forall(member(String,List),( String=term(Something)->write_the_term(Something);write(String)));
		write(List).

% ------------------------------------------------------------------
% public return_the_feedbacklist(+Errs, -FeedbackListSorted)
% This reads a list of errors and returns a sorted list of feedbacks
return_the_feedbacklist(Errs, NewFeedbackListSorted):-
	% find the appropriate feedback terms for the errors
	logging(feedbackHandler,'<feedbackHandler><RETURN_THE_FEEDBACKLIST>Errs=~p',[Errs]),
  	findall(feedback(Priority,Err,Where,What,How),(member(Err,Errs), feedback(Priority, Err, Where,What,How)),FeedbackList),
	logging(feedbackHandler,'<feedbackHandler><RETURN_THE_FEEDBACKLIST>FeedbackList=~p',[FeedbackList]),

	% sort the feedbacks according to their priority. The dubplicated feedbacks will be removed
	predsort(compareFeedback,FeedbackList, FeedbackListSorted),
	logging(feedbackHandler,'<feedbackHandler><RETURN_THE_FEEDBACKLIST>FeedbackListSorted=~p',[FeedbackListSorted]),
        (   (   member(Feedback, FeedbackListSorted), 
	    Feedback=feedback(_,error(_,typify_errors),_,_,_), 
	    length(FeedbackListSorted, Length),
	    Length>1) ->delete(FeedbackListSorted,Feedback,NewFeedbackListSorted);
	NewFeedbackListSorted=FeedbackListSorted
	).
	
%= private compareFeedback(?, +feedback(_), +feedback(_))
% this compare the priority of the feedbacks.
compareList(<, [_,Y1], [_,Y2]):-
	Y1<Y2.
sortList(ListIn, ListSorted):-predsort(compareList,ListIn,ListSorted).

compareFeedback(<, feedback(X,error(PosX,_),_,_,_), feedback(Y,error(PosY,_),_,_,_)):-
   X<Y;
   (X=Y, compareErrorPos(<, PosX,PosY)).
compareFeedback(>, feedback(X,error(PosX,_),_,_,_), feedback(Y,error(PosY,_),_,_,_)):-
	X>Y;
	(   X=Y, compareErrorPos(<,PosY,PosX)).
compareFeedback(=, feedback(X,error(PosX,_),_,_,_), feedback(Y,error(PosY,_),_,_,_)):-
	(   X=Y, not(compareErrorPos(<,PosX,PosY)),  not(compareErrorPos(<,PosY,PosX))).

%= private compareErrorPos(?,+Error1, +Error2)
% [X1,Y1]> [X2,Y2] 
% if 1) Y1>Y2
%    2) Y1=Y2, X1>X2
% [X1]>[X2] if X1>X2
% [X1,Y1]>[_]
% [_]> [] 
compareErrorPos(<,Error1, Error2):-
	(   Error1=[], Error2=[_]);
	(   Error1=[_], Error2=[X1,Y1]);
	(   Error1=[X1], Error2=[X2], X1<X2);
	(   Error1=[X1,Y1], Error2=[X2,Y2], (Y1<Y2;(Y1=Y2,X1<X2))).

% --------------------------------------------------------------------
%=public locate_error_in_term(+ErrorList,+Query)
% an method to locate the errors occurred in a query
% Error has following representations 
% error([1], unmatched_set2(buch('<var>'('Signatur'), '<a-var>'('_225')
% error([5, 1], unmatchable(5, nein, 5, '<a-var>'('_224'))
%
% Term is a query that has been processed with normalform:consumeAtom1
locate_error_in_term(Errs,Term):-
	% change the Term to a list by normalizing
	%normalform:normalizeBody(Term,SubgoalListWithVar,[]),
	findall(Subgoal, (member(SubgoalWithVar,Term), eliminateVarIndInTerm(SubgoalWithVar, Subgoal)),
SubgoalList),
	logging(feedbackHandler,'<feedbackHandler><LOCATE_ERROR_IN_TERM>TermNL=~p',[SubgoalList]),
	logging(feedbackHandler,'<feedbackHandler><LOCATE_ERROR_IN_TERM>Errs=~p',[Errs]),
	processSubgoalList(SubgoalList,Errs,SubgoalBoldList),
	write(SubgoalBoldList).


%= private processSubgoalList(+TermNL,+Errs,-TermBoldList)
% Errs is a list of error position [SubgoalPos, ArgPos].
% This change TermNL to TermBoldList with bold written errors.
processSubgoalList(SubgoalList,[],SubgoalList).
processSubgoalList(SubgoalList, [H|Rest], SubgoalBoldList):-
	write_error_in_bold(SubgoalList,H,TermBoldListTmp),
	processSubgoalList(TermBoldListTmp, Rest, SubgoalBoldList).


%= private contextError(+TermList,+ErrorPos,-NewTermList)
% This find the error in a subgoal and then highlights it
write_error_in_bold(SubgoalList,Error, NewSubgoalList):-
	logging(feedbackHandler,'<feedbackHandler><WRITE_ERROR_IN_BOLD>SubgoalList=~p',[SubgoalList]),
	errorPos(Error,SubgoalPos,ArgPos),
	logging(feedbackHandler,'<feedbackHandler><WRITE_ERROR_IN_BOLD>SubgoalPos=~p, ArgPos=~p',[SubgoalPos, ArgPos]),
	% select the SubgoalPos from the TermList
	n_tes(SubgoalPos,SubgoalList,Subgoal),
	logging(feedbackHandler,'<feedbackHandler><WRITE_ERROR_IN_BOLD>Subgoal=~p',[Subgoal]),
	(   nonvar(ArgPos), ArgPos=\=0)->
	(   % select the ArgPos from Subgoal
	    arg(ArgPos, Subgoal, Context),
            Subgoal =.. TermList,
	    logging(feedbackHandler,'<feedbackHandler><WRITE_ERROR_IN_BOLD>Context=~p',[Context]),
	    NewArgPos is ArgPos + 1,
            % we want to write the error context in bold
            (   atomic(Context)->
	        ( logging(feedbackHandler,'<feedbackHandler><WRITE_ERROR_IN_BOLD>Context is atomic=~p',[Context]),
	          atom_concat('<b>', Context, ContextTemp),
                  atom_concat(ContextTemp, '</b>', BoldContext)
	        );
	     
	        (logging(feedbackHandler,'<feedbackHandler><WRITE_ERROR_IN_BOLD>Context is not atomic =~p',[Context]),
                 term_to_atom(Context,ContextAtom),
		 logging(feedbackHandler,'<feedbackHandler><WRITE_ERROR_IN_BOLD>Context is converted to atom =~p',[ContextAtom]),
		 atom_concat('<b>', ContextAtom, ContextTemp),
	         atom_concat(ContextTemp, '</b>', BoldContext)
	        )),
    			  
	   logging(feedbackHandler,'<feedbackHandler><WRITE_ERROR_IN_BOLD>BoldContext=~p',[BoldContext]),     
           % and this bolded error context should be highlgihted within the subgoal
           replace_n_element(NewArgPos,TermList,BoldContext,NewTermList),
           NewSubgoal =.. NewTermList,
           logging(feedbackHandler,'<feedbackHandler><WRITE_ERROR_IN_BOLD>NewSubgoal=~p',[NewSubgoal]),
           % reconstruct the query
           replace_n_element(SubgoalPos, SubgoalList, NewSubgoal, NewSubgoalList)
	);
	(   
	    NewSubgoalList=SubgoalList,
	    logging(feedbackHandler,'<feedbackHandler><WRITE_ERROR_IN_BOLD>NewSubgoalList=~p',[NewSubgoalList])
	).
       

%= private errorPos(+Error,-ArgPos,-SubgoalPos)
% this extracts the error position from an error term
errorPos(Error,SubgoalPos,ArgPos):-
	% get the position of error
	arg(1,Error,ErrorPos),
	(   ErrorPos=[ArgPos,SubgoalPos];
	ErrorPos=[SubgoalPos]),
	logging(feedbackHandler,'<feedbackHandler><ERROR_POS>SubgoalPos=~p',[SubgoalPos]),
	logging(feedbackHandler,'<feedbackHandler><ERROR_POS>ArgPos=~p',[ArgPos]).

% -----------------------------------------------------------------------
	
%%
%% Reproducing the exercise description with emphasised parts.
%%

% erzeugt einen Ausgabetext, in dem die fehlerhaften Stellen im
% Fettdruck markiert sind
% alt: generate_text(XML_text,Fehlerliste,Fehlertext_mit_Bold_Markup)
% neu: generate_text(XML_text,Fehlerliste,Erledigte_Fehler,Unerledigte_Fehler,
%   Fehlertext_mit_Bold_Markup)

:-declare_struct(gt,[erledigt,unerledigt,text]).
%% gt(Erledigt,Unerledigt,Text)
%% Erledigt   -- List of error terms for which we have generated feedback
%% Unerledigt -- List of error terms we have left
%% Text       -- partial list of the text used for feedback

%=select_errorterm(Term) (*)
% take one error term from the list
select_errorterm(Term)-->
	modify(gt^unerledigt,Unerledigt0,Unerledigt1),
	{select(Term,Unerledigt0,Unerledigt1)},
	pushf(gt^erledigt,Term).

%=add_to_text(Tokens) (*)
% add the tokens to the feedback text
% we have to use popf here so that the list
% is built starting from the beginning
add_to_text([])-->[].
add_to_text([Text|Rest])-->
	popf(gt^text,Text),
	add_to_text(Rest).
	
%= generate_match(Attrs) (*)
% either removes a matching error from the list of error terms
% and succeeds
% or simply fails if there is no error term matching the description
generate_match([functor=Functor])-->    
	select_errorterm(error([_],unmatched_set2(Term))),
	{functor(Term,Functor,_)}.

generate_match([var=VarName])-->
	select_errorterm(error(_,unmatchable(_,_,_,'<var>'(VarName)))).

generate_match([constant=Constant])-->
	select_errorterm(error(_,unmatchable(_,_,_,Constant0))),
	{(Constant=Constant0;
	    number(Constant0),atom_number(Constant,Constant0))}.
generate_match([constant=Constant])-->
	select_errorterm(error(_,misspelled(_,Constant))).

generate_text([])-->[].
generate_text([Atom|Rest])-->
	{atom(Atom)},
	add_to_text([Atom]),
	generate_text(Rest).

generate_text([element(match,Attrs,Content)|Rest])-->
	(generate_match(Attrs)->
	% found the an error term for this functor
	    add_to_text(['<b>']),
	    generate_text(Content),
	    add_to_text(['</b>']);
        % no such error term
	    generate_text(Content)),
	% go on with the rest
	generate_text(Rest).

generate_text([element(Elm,Attrs,Content)|Rest])-->
	{print_message(warning,format("Unknown Element ~w ~w",[Elm,Attrs]))},
	generate_text(Content),
	generate_text(Rest).
