% Author: Yannick Versley
% Modified by: $Id: ids.pl,v 1.22 2004/03/29 07:41:55 nthinh Exp $

:- module(ids,[error/4,state2/4,
	       position/4, with_position/4,
	       expect/5,expect1/5,
	       add_to_map/4, add_to_bimap/4,
	       match_lists/7, match_lists_/7, match_lists1/8, match_lists_1/8,
	       match_bags/6, match_bags/7,  match_bags1/7,
	       apply_to_list/4, apply_to_list1/4,
	       ids_search/7,ids_search/5,ids_search/3]).
:- use_module(struct).
:- use_module(logging).
:- module_transparent 
         ids_search/3, ids_search/5, ids_search/7, ids_search/8,
	 expect/5, expect1/5,
	 match_lists/7, match_lists_/7,
	 match_bags/6, match_bags/7,
	 apply_to_list/4, with_position/4.
%%
%% iterative deepening search
%%

% points: 
% position is the postion of goal/subgoal
% errors is a list of errors
% state2 is the state of the object that ids is operating on.
:-declare_struct(ids_state,[points,position,
			    errors,state2]).

%=error(Erklrung,Strafe) (*)
% falls noch mindestens <i>Strafe</i>
% Fehlerpunkte ber sind, werden diese abgezogen;
% wenn nicht, scheitert der Aufruf.
error(Reason, Penalty,
      ids_state(N,Pos,Errors,State2),
      ids_state(N1,Pos,[error(Pos,Reason)|Errors],State2)):-
	N1 is N-Penalty,
	N1>=0,
	logging(ids,'<error>Error with Reason=~w, Penalty=~w, and new state State2=~w',[Reason,Penalty,State2]).

%=position(PosOld,PosNew) (*)
% holt und setzt die Positionsinformation
position(PosOld,PosNew)-->
	{logging(ids,'<position>PosOld=~w, PosNew=~w',[PosOld,PosNew])},
	modify(ids_state^position,PosOld,PosNew).

with_position(PosArg,Query)-->
	{logging(ids,'<with_position>Position PosArg=~w',[PosArg])},
	pushf(ids_state^position,PosArg),
	{logging(ids,'<with_position>Calling Query=~w',[Query])},
	call(Query),
	popf(ids_state^position,_).

%=match_lists_(List1,List2,MatchPred,ErrTerm,Pen)
% matches two lists without considering misplaced ordering
match_lists_(L1,L2,MP,ET,Pen)-->
	{logging(ids,'<match_lists>apply MatchPred=~w on List1=~w and List2=~w',[MP,L1,L2]),
	 logging(ids,'<match_lists>with ErrorTerm=~w and Penalty=~w',[ET,Pen])
	},
	with_position(0,match_lists_1(0,L1,L2,MP,ET,Pen)).


%=match_lists(List1,List2,MatchPred,ErrTerm,Penalty) (*)
% wendet <i>MatchPred</i> auf die entsprechenden
% Elemente von <i>List1</i> und <i>List2</i> an.
% Soll sich auch um vergessene/vertauschte Listenelemente kmmern.

%%% Problem: Nummerierung von List1 vs List2
%%%	     Position soll aus List1 kommen, Stelle in List2
%%%	     muss so mitgefhrt werden
match_lists(L1,L2,MP,ET,Pen)-->
	{logging(ids,'<match_lists>apply MatchPred=~w on List1=~w and List2=~w',[MP,L1,L2]),
	 logging(ids,'<match_lists>with ErrorTerm=~w and Penalty=~w',[ET,Pen])
	},
	with_position(0,match_lists1(0,L1,L2,MP,ET,Pen)).
match_lists(_,_,_,_,_)-->
	error(lists_do_not_match,30).


match_lists1(_,[],[],_,_,_)-->[].
match_lists1(M,[E1|Rest1],[E2|Rest2],MatchPred,ErrTerm,Penalty)-->
	{logging(ids,'<match_list1>try matching the first elements E1=~w, E2=~w with Rest1=~w, Rest2=~w',[E1,E2,Rest1,Rest2]),
	 logging(ids,'<match_list1>with MatchPred=~w, ErrorTerm=~w, Penalty=~w',[MatchPred,ErrTerm,Penalty])
	},
	position([N|Pos0],[N1|Pos0]),
	{succ(N,N1),
	 succ(M,M1),
	 ErrTerm=..ETL0,
	 append(ETL0,[N1,E1,M1,E2],ETL1),
	 RealErrTerm=..ETL1
	},
	expect(call(MatchPred,E1,E2),RealErrTerm,Penalty),
	match_lists1(M1,Rest1,Rest2,MatchPred,ErrTerm,Penalty).
match_lists1(M,List1,[E2|Rest2],MP,ET,Pen)-->
	{logging(ids,'<match_list1>try matching the List1=~w and E2=~w,Rest2=~w',[List1,E2,Rest2]),
	 logging(ids,'<match_list1>with MatchPred=~w, ErrorTerm=~w, Penalty=~w',[MP,ET,Pen])
	},
	position([N|Pos0],[N|Pos0]),
	{succ(N,N1),succ(M,M1)},
	error(forgotten_list_element(N1,E2),5),
	
	match_lists1(M1,List1,Rest2,MP,ET,Pen).
% ein Element aus List1 entspricht einem spteren Element aus List2
match_lists1(M,[E1|Rest1],List2,MP,ET,Pen)-->
	position([N|Pos0],[N1|Pos0]),
	{succ(N,N1),
	 append(Before2,[E2|After2],List2),
	 length(Before2,NB),
	 N2 is N1+NB,
	 M1 is N+NB,
	 succ(M1,M2),
	 length(Before1,NB),
	 append(Before1,After1,Rest1),
	 ET=..ETL0,
	 append(ETL0,[N1,E1,M1,E2],ETL1),
	 RealErrTerm=..ETL1
	},
	expect(call(MP,E1,E2),RealErrTerm,Pen),
	error(misplaced_list_element(N1,E1,M2,E2),5),
	match_lists1(M,Before1,Before2,MP,ET,Pen),
	position([_|Pos0],[N2|Pos0]),
	match_lists1(M2,After1,After2,MP,ET,Pen).
% ein Element aus List2 entspricht einem spteren Element aus List1
match_lists1(M,List1,[E2|Rest2],MP,ET,Pen)-->
	position([N|Pos0],[N|Pos0]),
	{succ(M,M1),
	 append(Before1,[E1|After1],List1),
	 length(Before1,NB),
	 N1 is N+NB,
	 M2 is M1+NB,
	 succ(N1,N2),
	 length(Before1,NB),
	 append(Before2,After2,Rest2),
	 ET=..ETL0,
	 append(ETL0,[N1,E1,M1,E2],ETL1),
	 RealErrTerm=..ETL1
	},
	expect(call(MP,E1,E2),RealErrTerm,Pen),
	error(misplaced_list_element(N1,E1,M1,E2),5),
	match_lists1(M1,Before1,Before2,MP,ET,Pen),
	position([_|Pos0],[N2|Pos0]),
	match_lists1(M2,After1,After2,MP,ET,Pen).
match_lists1(_,[_|_],[],_,_,_)-->
	error(superfluous_list_elements,10).
match_lists_1(_,[],[],_,_,_)-->[].
match_lists_1(M,[E1|Rest1],[E2|Rest2],MatchPred,ErrTerm,Penalty)-->
	position([N|Pos0],[N1|Pos0]),
	{succ(N,N1),
	 succ(M,M1),
	 ErrTerm=..ETL0,
	 append(ETL0,[N1,E1,M1,E2],ETL1),
	 RealErrTerm=..ETL1
	},
	expect(call(MatchPred,E1,E2),RealErrTerm,Penalty),
	match_lists_1(M1,Rest1,Rest2,MatchPred,ErrTerm,Penalty).

%=match_bags(List1,List2,PosMap,MatchPred) (*)
% applies <i>MatchPred</i> on elements from <i>List1</i>
% and <i>List2</i>;
% The mapping of positions is saved in PosMap
% in contrast to match_lists, order does not matter
match_bags(L1,L2,MP,Pen)-->
	match_bags(L1,L2,_,MP,Pen).
match_bags(L1,L2,PosMap,MP,Pen)-->
	{findall(N-SG,nth0(N,L2,SG),L2a)},
	with_position(0,match_bags1(L1,L2a,PosMap,MP,Pen)).

match_bags1([],[],[],_,_)-->[].
match_bags1([E1|Rest],L2,[N=M|RestMap],MatchPred,Pen)-->
	position([N|Pos0],[N1|Pos0]),
	{
	 succ(N,N1),
	 select(M-E2,L2,Rest2)
	},
	call(MatchPred,E1,E2),
	match_bags1(Rest,Rest2,RestMap,MatchPred,Pen).
match_bags1([E1|Rest],L2,PMap,MP,Pen)-->
	error(unmatched_set1(E1),Pen),
	match_bags1(Rest,L2,PMap,MP,Pen).
match_bags1([],[_-E2|Rest],PMap,MP,Pen)-->
	error(unmatched_set2(E2),Pen),
	match_bags1([],Rest,PMap,MP,Pen).

%=apply_to_list(List,MatchPred,ErrTerm,Penalty)
apply_to_list(List,MatchPred)-->
	{logging(ids,'<apply_to_list>List=~w, MatchPred=~w',[List,MatchPred])},
	with_position(0,apply_to_list1(List,MatchPred)).
apply_to_list1([],_)-->
	[],
	{logging(ids,'<apply_to_list1a>applied to empty list',[])}.
apply_to_list1([E|Rest],Pred)-->
	position([N|Pos0],[N1|Pos0]),
	{succ(N,N1)},
	{logging(ids,'<apply_to_list1b>try to apply Pred=~w to element E=~w',[Pred,E])},
	call(Pred,E),
	apply_to_list1(Rest,Pred).
	

%=state2(StateOld,StateNew) (*)
state2(StateOld,StateNew)-->
	modify(ids_state^state2,StateOld,StateNew).

%=expect(DCG_Aufruf,Erklrung,Strafe) (*)
% falls <i>DCG_Aufruf</i> gelingt, so wird mit dessen Ergebnis
% fortgefahren, ansonsten wird dies als Fehler gewertet.
expect(What,Reason,Penalty,ListIn,ListOut):-
	logging(ids,'<expect>try to call What=~w with ListIn=~w, ListOut=~w',[What,ListIn,ListOut]),
	call(What,ListIn,List1) *-> List1=ListOut ;
	error(Reason,Penalty,ListIn,ListOut),
	logging(ids,'<expect>not successful, error with Reason=~w and Penalty=~w',[Reason,Penalty]),
	logging(ids,'<expect>ListIn=~w, ListOut=~w',[ListIn,ListOut]).

%=expect1(Aufruf,Erklrung,Strafe) (*)
% falls <i>Aufruf</i> gelingt, so wird fortgefahren, ansonsten
% wird dies als Fehler gewertet.
expect1(What,Reason,Penalty,ListIn,ListOut):-
	call(What) *-> 
	    (	ListIn=ListOut, 
	       logging(ids,'<expect1>called What=~w with ListIn=~w, ListOut=~w',[What,ListIn,ListOut])
	    );
	    (	error(Reason,Penalty,ListIn,ListOut),
		logging(ids,'<expect1>not successful, error with Reason=~w and Penalty=~w',[Reason,Penalty]),
		logging(ids,'<expect1>ListIn=~w, ListOut=~w',[ListIn,ListOut])
	    ).


%=add_to_map(Key,Value,MapOld,MapNew)
add_to_map(Key,Value,MapOld,MapNew):-
	logging(ids,'<add_to_map>Input: Key=~w, Value=~w, MapOld=~w',[Key,Value,MapOld]),
	member(Key=SomeValue,MapOld)->
	   unify_with_occurs_check(Value,SomeValue),
	   MapNew=MapOld;
	   MapNew=[Key=Value|MapOld],
	logging(ids,'<add_to_map>Output: MapNew=~w',[MapNew]).

%=add_to_bimap(Key,Value,MapOld,MapNew)
add_to_bimap(Key,Value,MapOld,MapNew):-
	member(Key=Value,MapOld),
	MapNew=MapOld;
	not(member(Key=_,MapOld)),
	not(member(_=Value,MapOld)),
	MapNew=[Key=Value|MapOld].

% Find the optimal derivation for Query with the initial state
% of State0.
% +Query   - the query to execute
% +State0  - the initial state
% -Errs    - the list of errors
ids_search(Query,State0,Errs):-
	ids_search(Query,[],State0,_,0,50,Errs).

ids_search(Query,State0,State1,MaxPenalty,Errs):-
	ids_search(Query,[],State0,State1,0,MaxPenalty,Errs).


% Query is the description of the search space.
% Pos0 
% State0 is the beginning state of the object to be operated
% State1 is the last state
% MinPenalty 
% MaxPenalty is the max depth bound
% Errors is the solution path of error terms.
ids_search(Query,Pos0,State0,State1,
	   MinPenalty,MaxPenalty,
	   Errors):-
	logging(ids, '<ids_search1> Query=~w, Pos0=~w, State0=~w, State1=~w, MinPenalty=~w, MaxPenalty=~w', 
	      [Query,Pos0,State0,State1,MinPenalty,MaxPenalty]),
	StartTime is cputime,
	ids_search(Query,Pos0,State0,State1,
		   MinPenalty,MaxPenalty,
		   Errors,StartTime).
       

ids_search(Query,Pos0,State0,State1,
	   MinPenalty,MaxPenalty,
	   Errors,StartTime):-
	logging(ids, '<ids_search2> Query=~w, Pos0=~w, State0=~w, State1=~w, MinPenalty=~w, MaxPenalty=~w,Errors=~w, StartTime=~w', 
	                                              [Query,Pos0,State0,State1,MinPenalty,MaxPenalty,Errors,StartTime]),
	CurrentTime is cputime,
	TimeElapsed is CurrentTime-StartTime,
	logging(ids,'<ids_search2> Trying for MinPenalty=~w after ~w seconds',[MinPenalty,TimeElapsed]),
	% if the TimeElapsed exceeds 0.4 sec than search will be aborted.
	TimeElapsed>0.4 ->
	%TimeElapsed>1 ->
	logging(ids,'search aborted...',[]),!,fail;
	
	% find a list of solutions with cost less than MinPenalty
	% 
	findall(ErrorCost-result(Errs,State1,Query),
		(logging(ids,'<ids_search2>Calling Query=~w, with first ids_state(_,_,_,~w), second ids_state(_,_,_,~w)',[Query,State0,State1]),
		call(Query,ids_state(MinPenalty,Pos0,[],State0),ids_state(Points,_,Errs,State1)),ErrorCost is MinPenalty-Points),
		List),
	logging(ids,'<ids_search2>List of solutions=~w',[List]),
	% if there is one solution in the list, return the first element (best solution) of the sorted list.
	List\=[]->
	     keysort(List,List1), % sort the list of solutions with increasing ErrorCost.
	     List1=[Best-_|_],
	     % sollte die Liste abschneiden...
	     member(Best-result(Errors,State1,Query),List1);
	MinPenalty<MaxPenalty,
	NewPenalty is MinPenalty+5,
	logging(ids,'<ids_search2>next attempt with NewPenalty=~w',[NewPenalty]),
	ids_search(Query,Pos0,State0,State1,NewPenalty,MaxPenalty,
		  Errors,StartTime).






