% Created by Yannick Versley.
% Modified: $Id: infer_types.pl,v 1.25 2004/04/13 14:52:24 nthinh Exp $ 


:- module(infer_types,[get_type_from_imm/4, typed_pred/3,
		       get_types_from_exprs/3,
		       typify_query/2,
		       typify_query/3,
		       typify_pred/3,
		       typify_arithmetic/3,
		      current_domain/1, set_domain/1]).

:- use_module(normalform).
:- use_module(ids).
:- use_module(logging).
:- multifile range_type/3.
:- multifile enum_type/2.
:- multifile pred_type/2.
:- dynamic range_type/3.
:- dynamic enum_type/2.
:- op(900,fx,type_decl).

:- op(700,xfx,=>).
:- op(700,xfx,<=).
:- op(700,fx, < ).
:- op(700, xf, <).
:- op(700,fx, > ).
:- op(700,xf, > ).


:- multifile user:term_expansion/2.

:- declare_struct(types,[var,pred]).

:- thread_local current_domain/1.
:- retractall(current_domain(_)).

set_domain(Dom):-
	retractall(current_domain(_)),
	assert(current_domain(Dom)),
	debug(types,'<set_domain>current domain is ~w',[Dom]),
	set_logfile(Dom).

builtin(append(_,_,_)).
builtin(_\=_).
builtin(_=_).
builtin(member(_,_)).

%range_type(jahr,1900,2050).
%range_type(kalenderwoche,1,53).
%enum_type(peano,[0,s(peano)]).
%enum_type(janein,[ja,nein]).
%enum_type(lst(A),[[A|lst(A)],[]]).

%% Macros for type declaration
%%
%% these are used in the <domain>_types.pl files
%% and have to be dependent on the currently selected domain
%%
%% see bibliothek_typen.pl for examples

%=range_decl(Typname, Min, Max)
% Typname ist als Zahlenbereich von Min bis Max deklariert
user:term_expansion(range_decl(Type,Lower,Upper),
		 (infer_types:range_type(Type,Lower,Upper)
		 :-current_domain(Dom))):-
	current_domain(Dom).
%=enum_decl(Typname,Konstruktoren)
% Typname ist mit ein Typ, dessen Elemente durch
% die genannten Funktoren realisiert sind.
user:term_expansion(enum_decl(Type,Values),
		 (infer_types:enum_type(Type,Values)
		 :-current_domain(Dom))):-
	current_domain(Dom).
%=type_decl(Template)
% Template has the same functor/arity as the predicate
% it describes;
% the arguments are the types of the arguments for the predicate.
user:term_expansion(type_decl(Template),
		 (infer_types:pred_type(Pred,ArgTypes)
		 :-current_domain(Dom))):-
	current_domain(Dom),!,
	Template=..[Pred|ArgTypes].
user:term_expansion(type_decl(Template),
		 (infer_types:pred_type(Pred,ArgTypes))):-
	not(current_domain(_)),
	Template=..[Pred|ArgTypes].

number_type(number).
number_type(RangeType):-range_type(RangeType,_,_).

more_special(RangeType,number):-range_type(RangeType,_,_).

%=compare_op(Op)
% Op ist ein Vergleichsoperator fr Zahlen
compare_op(<).
compare_op(>).
compare_op(=<).
compare_op(>=).
%compare_op(\=).
%compare_op(==).
%compare_op(\==).


faulty_compare_op(=>,>=).
faulty_compare_op(<=,=<).

vdashlike(:).
vdashlike(->).
vdashlike(:=).


%=type_decl pred(Typ1,...,Typn)
% Die Argumente von <i>pred</i> haben die Typen
% Typ<i>i</i>
type_decl A\=A.
type_decl A = A.
type_decl member(A,lst(A)).
type_decl reverse(lst(A),lst(A)).
type_decl append(lst(A),lst(A),lst(A)).
type_decl sort(lst(A),lst(A)).

unify_types(Type1,Type2)-->{Type1=Type2}.
unify_types(Type1,Type2)-->
	{ once(A\=B; not(ground(A)); not(ground(B)))},
	error(should_be_same(Type1,Type2),5).


% check if the student use prefix op for >, <

add_var(Varname,Type)-->
	modify(ids_state^state2-types^var,VarMap0,VarMap1),
	{add_to_map(Varname,Type,VarMap0,VarMap1),
	logging(types,'<add_var1>added Varname=~w, Type=~w, changed VarMap0=~w to VarMap1=~w',[Varname,Type,VarMap0,VarMap1])}.
add_var(Varname,Type)-->
	pushf(ids_state^state2-types^var,Varname=Type),
	error(conflicting_type(Varname,Type),5),
	{logging(types,'<add_var2>added Varname=~w, Type=~w',[Varname,Type])}.

%=get_type_from_imm(Term,Typ) (*)
% <i>Term</i> is a term of the type <i>Typ</i>
% unifies Typ with a type that could be the one of Term
% Variables are added to a map so they have to have the same type
% throughout multiple subgoals
get_type_from_imm('<var>'(Varname),Type)-->
/**	{ downcase_atom(Varname,Downcased),
	  (range_type(Downcased,_,_); enum_type(Downcased,_))->
	  Type=Downcased; true },*/
	{logging(types,'<get_type_from_imm1>Input Varname=~w',[Varname])},
	add_var(Varname,Type),
	{logging(types,'<get_type_from_imm1>Output Type=~w',[Type])}.

get_type_from_imm('<a-var>'(_),_)-->
	[],
	{logging(types,'<get_type_from_imm2>Term is a anonymous variable',[])}.

get_type_from_imm(X,RangeType)-->
	{logging(types,'<get_type_from_imm3>Input X=~w',[X])},
	{ range_type(RangeType,Min,Max),
	  number(X),
	  X>=Min, X =< Max},
	{logging(types,'<get_type_from_imm3>Input X=~w has RangeType=~w',[X,RangeType])}
	.
get_type_from_imm(X,EnumType)-->
	{logging(types,'<get_type_from_imm4>Input X=~w',[X])},
	{ once((X=..[Functor|Args],
		enum_type(EnumType,Constructors),
		
		member(Cons,Constructors),
		Cons=..[Functor|ArgTypes])) ,
	  logging(types,'<get_type_from_imm4>enum_type=~w has Constructors=~w',[EnumType,Constructors]),
	  logging(types,'<get_type_from_imm4>Cons=~w',[Cons])
	},
	{logging(types,'<get_type_from_imm4>EnumType=~w',[EnumType])},
	match_typelist(Args,ArgTypes).

get_type_from_imm(Atom,TypeX)-->
	% Atom is assumed to be a variable
	{logging(types,'<get_type_from_imm5>Input Atom=~w',[Atom])},
	error(atom_as_var(Atom),10),
	{atom(Atom),
	 upcase_atom(Atom,ATOM),
	 not((enum_type(_EnumType,Constructors),
	      member(Atom,Constructors),
	      logging(types,'<get_type_from_imm5>Atom=~w is member of enum_type=~w',[Atom,Constructors])
	     ))
	},
	get_type_from_imm('<var>'(ATOM),TypeX),
	{logging(types,'<get_type_from_imm5>Atom=~w has TypeX=~w',[Atom,TypeX])}.

get_type_from_imm(X,number)-->
	{logging(types,'<get_type_from_imm6>Input Varname=~w',[X])},
	{number(X),
	logging(types,'<get_type_from_imm6>Input Varname=~w is of type number',[])
	}.

get_type_from_imm(Term,NumberType)-->
	{logging(types,'<get_type_from_imm7>Input Term=~w',[Term])},
	error(inline_compare(Term),5),
	{ number_type(NumberType),
	  Term =.. [CompareOp,Var,Number]},
	{logging(types,'<get_type_from_imm7>Term is seperated in CompareOp=~w, Var=~w, Number=~w',[CompareOp,Var,Number])},
	({compare_op(CompareOp)};{faulty_compare_op(CompareOp,RealCompareOp)},
	   error(use_X_not_Y(RealCompareOp,CompareOp),0)),
	  {Var='<var>'(_)},
	get_type_from_imm(Var,NumberType),
	get_type_from_imm(Number,NumberType),
	{logging(types,'<get_type_from_imm7>Term=~w has NumberType=~w',[Term,NumberType])}.

% this checks the prefix op <,>
get_type_from_imm(Term,NumberType)-->
	{logging(types,'<get_type_from_imm8>Input Term=~w',[Term])},
	error(inline_compare(Term),5),
	{ number_type(NumberType),
	  Term =.. [CompareOp,Number]},
	{logging(types,'<get_type_from_imm8>Term is seperated in CompareOp=~w, Number=~w',[CompareOp,Number])},
	({compare_op(CompareOp)},
	   error(not_inside_comp(CompareOp),5)),
	get_type_from_imm(Number,NumberType),
	{logging(types,'<get_type_from_imm8>Term=~w has NumberType=~w',[Term,NumberType])}
	.

get_type_from_imm(Term,NumberType)-->
	{logging(types,'<get_type_from_imm9>Input Term=~w',[Term])},
	error(inline_arithmetic(Term),10),
	typify_arithmetic(Term),
	{ number_type(NumberType) },
	{logging(types,'<get_type_from_imm9>Term=~w has NumberType=~w with error inline_arithmetic.',[Term,NumberType])}.

%=typify_arithmetic(Term) (*)
% <i>Term</i> ist ein arithmetischer Ausdruck,
% wie er bei =:=, is etc. vorkommt.
typify_arithmetic(X)-->
	{logging(types,'<typify_arithmetic1>Input X=~w',[X])},
	{number(X),!},
	{logging(types,'<typify_arithmetic1>Input X=~w is a number',[X])}.

typify_arithmetic(X)-->
	{logging(types,'<typify_arithmetic2>Input X=~w',[X])},
	{
	 not(atomic(X)),
	 X=..[_|Args]
	},
	% current_arithmetic_function is not defined yet ???
	expect1(current_arithmetic_function(X),no_arithmetic_function(X),5),
	{logging(types,'<typify_arithmetic2>Input X=~w is not atomic',[X])},
	apply_to_list(Args,typify_arithmetic),
	{logging(types,'<typify_arithmetic2>applied typify_arithmetic to the arguments Args=~w of Input X=~w',[Args,X])}.

typify_arithmetic(X)-->
	{logging(types,'<typify_arithmetic3>Input X=~w',[X])},
	get_type_from_imm(X,NumType),
	{logging(types,'<typify_arithmetic3>X=~w has type=~w',[X,NumType])},
	expect1(number_type(NumType),
		must_be_number(X),
		10),
	{logging(types,'<typify_arithmetic3>Input X=~w has number_type=~w',[X,NumType])}.


%=typed_pred(Term) (*)
% Term ist ein Aufruf, dessen Argumenten Typen zugewiesen werden.
typed_pred(Term)-->
	{logging(types,'<typed_pred1>Input Term=~w',[Term])},
	{ Term=..[Functor|Args],
	  pred_type(Functor,ArgTypes)},
	match_typelist(Args,ArgTypes),
	{logging(types,'<typed_pred1>Output Term=~w has been operated to Functor=~w and Args=~w, has ArgTypes=~w ',[Term,Functor,Args,ArgTypes])}.

/*typed_pred(Term)-->
	{ Term=..[Functor,Arg1,Arg2],
	  compare_op(Functor)},
	get_type_from_imm(Arg1,Type1),
	get_type_from_imm(Arg2,Type2),
	unify_types(Type1,Type2),
	(
	expect1(number_type(Type1),must_be_number(Type1),5);
	expect1(number_type(Type2),must_be_number(Type2),5)
	).
*/
typed_pred(Term)-->
	{logging(types,'<typed_pred2>Input Term=~w',[Term])},
	{Term=..[Functor,Arg1,Arg2],
	  compare_op(Functor)},!,
	typify_arithmetic(Arg1),
	typify_arithmetic(Arg2),
	{logging(types,'<typed_pred2>Output Term=~w has benn operated to Functor=~w and Arg1=~w and Arg2=~w ',[Term,Functor,Arg1,Arg2])}.

typed_pred(Term)-->
	{logging(types,'<typed_pred3>Input Term=~w',[Term])},
	{Term=..[Functor,Arg1,Arg2],
	  faulty_compare_op(Functor,RealFunctor)},!,
	error(use_X_not_Y(RealFunctor,Functor),0),
	typify_arithmetic(Arg1),
	typify_arithmetic(Arg2),
	{logging(types,'<typed_pred3>Output Term=~w, Functor=~w schould be ~w; Arg1=~w and Arg2=~w ',[Term,Functor,RealFunctor,Arg1,Arg2])}.

% check if student uses prefix operator for <, >
typed_pred(Term)-->
	{logging(types,'<typed_pred4>Input Term=~w',[Term])},
	{ Term=..[Functor,Arg1],
	compare_op(Functor)
	},!,
	error(not_inside_comp(Functor),5),
	typify_arithmetic(Arg1),
	{logging(types,'<typed_pred4>Output Term=~w: Functor=~w and Arg1=~w with error',[Term,Functor,Arg1])}.

typed_pred(Var is Term)-->
	{logging(types,'<typed_pred5>Input Term=~w',[Term])},
	get_type_from_imm(Var,NumType),
	typify_arithmetic(Term),
	expect1(number_type(NumType),must_be_number(Var),5),
	{logging(types,'<typed_pred5>Output Var=~w is Term=~w and Var has type=~w',[Var,Term,NumType])}.

typed_pred(Term)-->
	{logging(types,'<typed_pred6>Input Term=~w',[Term])},
	% find the functor
	{ Term=..[Functor1|Args],
	  % find the types of the arguments ie. append(lst(A),lst(A),lst(A))
	  pred_type(Functor,ArgTypes),
	  dwim_match(Functor1,Functor)},
	error(misspelled_predicate(Functor1,Functor),5),
	match_typelist(Args,ArgTypes),
	{logging(types,'<typed_pred6>Output Term=~w: Functor=~w schould be ~w, Args=~w has ArgTypes',[Term,Functor,Functor1,Args,ArgTypes])}.

typed_pred(Term)-->
	{logging(types,'<typed_pred7>Input Term=~w',[Term])},
	{ Term=..[Functor|Args]},
	getf(ids_state^state2-types^pred,Pred),
	{member(Functor=ArgTypes,Pred)}, % Was bedeutet das???
	match_typelist(Args,ArgTypes),
	{logging(types,'<typed_pred7>Output Term=~w: Functor=~w is member of Pred=~w, Args=~w has ArgTypes=~w',[Term,Functor,Pred,Args,ArgTypes])}.
	
typed_pred(Term)-->
	{logging(types,'<typed_pred8>Input Term=~w',[Term])},
	{ Term=..[Functor|_] },
	error(unknown_predicate(Functor),30),
	{logging(types,'<typed_pred8>Output Term=~w: Functor=~w is unknown predicate',[Term,Functor])}.

%=match_typelist(Functor,Args,ArgTypes)
% sucht Typen fr Args
% berprft die Typen der einzelnen Argumente mittels ArgTypes.
match_typelist(Args,ArgTypes)-->
	{logging(types,'<match_typelist>types for Args=~w, ArgTypes=~w',[Args,ArgTypes])},
	match_lists(Args,ArgTypes,
		    get_type_from_imm,
		    type_expected, 10),
	{logging(types,'<match_typelist>Output: ArgTypes=~w',[ArgTypes])}.

%=get_types_from_exprs(Exprs)
% Expr ist eine Liste von Termen aus einer
% Prolog-Anfrage bzw. einem Klauselkrper,
% auf denen Typinferenz betrieben wird.
get_types_from_exprs(Exprs)-->
	{logging(types,'<get_types_from_exprs>Input Exprs=~w',[Exprs])},
	apply_to_list(Exprs,typed_pred).

typify_query(Query,Errors):-
	ids_search(get_types_from_exprs(Query),types([],[]),Errors).
typify_query(Query,Errors,VarTypes):-
	ids_search(get_types_from_exprs(Query),
		   types([],[]),types(VarTypes,_),50,Errors).


%=typify_pred(Clauses,PredTypes,Errors)
% Clauses are the clauses of the predicate(s)
%   transformed by normalform:normalize/2.
% PredTypes are the guessed types of the predicates
% Errors are the errors that occurred.
typify_pred(Clauses,PredTypes,Errors):-
	logging(types,'<typify_pred>Input Clauses =~w',[Clauses]),
	ids_search(get_types_from_preds(Clauses),
		   types([],[]),types(_,PredTypes),
		   50,Errors).
		  
get_types_from_preds([])-->[].
get_types_from_preds([Head-Body|Rest])-->
	{Head=..[F|Args], not(current_op(_,_,F))},
	{logging(types,'<types_from_preds>operated Head=~w into Functor F=~w and Args=~w',[Head,F,Args])},
% Woher hat man types^pred, ids_state^points, ^position, ^errors ???
	setf(ids_state^state2-types^var,[]),
	% holt Liste von Prdikatentypen [Functor=Argumententypen]
	getf(ids_state^state2-types^pred,Pred),
	{logging(types,'<types_from_preds>Pred=~w',[Pred])},
	getf(ids_state^points,Points),
	{logging(types,'<types_from_preds>Points=~w',[Points])},
	getf(ids_state^position,Position),
	{logging(types,'<types_from_preds>Position=~w',[Position])},
	getf(ids_state^errors,Errors),
	{logging(types,'<types_from_preds>Errors=~w',[Errors])},
		
	(
	{member(F=ArgTypes,Pred)}->match_typelist(Args,ArgTypes);
	% in der 1. Klausel existiert noch keine Argumententypen fr den Functor.
	% versuchen wir die Typen fr diese Argumente herauszufinden.
	match_lists_(Args,ArgTypes,
		     get_type_from_imm,type_expected,10),
	% modify und add_to_map  knnen durch die Prozedur: getf(Predin), add_to_map, setf(Predout) ersetzt werden.
	modify(ids_state^state2-types^pred,PredIn,PredOut),
	{add_to_map(F,ArgTypes,PredIn,PredOut),
	 logging(types,'<types_from_preds>Functor=~w has ArgTypes=~w, map PredIn=~w, map PredOut=~w',[F,ArgTypes,PredIn,PredOut])
	}
	),
	{logging(types,'<types_from_preds>begin get_types_from_exprs with Body=~w',[Body])},
	get_types_from_exprs(Body),
	{logging(types,'<types_from_preds>begin get_types_from_preds with Rest=~w',[Rest])},
	get_types_from_preds(Rest).

%% If there is an outmost operator that is not :-, there is almost certainly
%% something wrong.
%% A parser would allow us to do a better guess on what actually happened
%% (e.g. '.' instead of ',', ':' or ':=' instead of ':-' etc.
get_types_from_preds([Head-[]|Rest])-->
	{functor(Head,F,_), current_op(_,_,F)},
	error(wrong_head(Head,F),5),
	get_types_from_preds(Rest).

