:- module(session,[get_session/1, encode_url/2, encode_url2/2,
		   session_setValue/2, session_getValue/2,
		  retract_session/0, set_session/1,
		  set_session_from_request/1,
		  write_session_cookie/0]).
:- use_module(library(url)).
:- use_module(logging).

:- dynamic current_session/1.
% :- thread_local current_session/1.
:- dynamic session_var/3.

random_codes(0, []).
random_codes(N, [Code|Codes]):-
	succ(N1,N),
	Code is 0'a+random(26),
	random_codes(N1,Codes).

get_session(SID):-
	current_session(SID),!.
get_session(SID):-
	random_codes(16,Codes),
	atom_codes(SID,Codes),
	asserta(current_session(SID)).

set_session(SID):-
	retractall(current_session(_)),
	asserta(current_session(SID)).

retract_session:-
	retractall(current_session(_)).

encode_url(URL,EncodedURL):-
	current_session(_),!,
	atom_codes(URL,Codes),
	phrase(url:uri(http,Parts),Codes),
	encode_url2(Parts,EncodedURL).
	
encode_url(URL,URL).

encode_url2(Parts,EncodedURL):-
	(current_session(SID)->
	(select(search(Search),Parts,Parts0)->
	phrase(url:curi([search([sid=SID|Search])|Parts0]),Codes1);
	phrase(url:curi([search([sid=SID])|Parts]),Codes1));
	phrase(url:curi(Parts),Codes1)),
	atom_codes(EncodedURL,Codes1).

session_setValue(Key,Val):-
	get_session(SID),
	retractall(session_var(SID,Key,_)),
	assert(session_var(SID,Key,Val)).
session_getValue(Key,Val):-
	current_session(SID),
	session_var(SID,Key,Val).

set_session_from_request(Request):-
	memberchk(search(Search),Request),
	memberchk(sid=SID,Search),!,
	set_session(SID).

set_session_from_request(_):-
	retract_session.

write_session_cookie.
