% Created by Yannick Versley.
% Modified: $Id: struct.pl,v 1.5 2004/04/13 14:53:32 nthinh Exp $ 


:-module(struct,[getf/4,setf/4,modify/5,pushf/4,popf/4]).

:-multifile user:term_expansion/2.
:-multifile user:goal_expansion/2.
:-multifile getf/4, setf/4, modify/5.
:-discontiguous getf/4, setf/4, modify/5.

%% This is an implementation of a named product, like
%% struct in C or record in Pascal.
%% This could be done more efficiently with setarg,
%% but this would mean we would need to be _very_ cautious

%% Usage:
%%   :- declare_struct(foo,[bar,baz]).
%% declares a struct foo with members bar and baz
%% when used as implicit argument to a predicate in
%% DCG-syntax, the predicates getf, setf etc. can be used
%% to get and set values, for example with
%%   baz_to_bar-->getf(foo^baz,X),setf(foo^baz,X).
%% Place descriptors can be composed with -
%% for example, if the baz place contains another foo struct,
%% we can use
%%   get_bazbar(X)-->getf(foo^baz-foo^bar,X).
%% to access the bar component of the baz component.
%%
%% Precautions:
%% When used with an undefined place descriptor,
%% the predicates simply fail.
%% For problems with the macro expansion, see below.

expand_inline(getf).
expand_inline(setf).
expand_inline(pushf).
expand_inline(popf).

getf(Place,Thing,All,All):-
	modify(Place,Thing,Thing,All,All).
setf(Place,Thing,All0,All1):-
	modify(Place,_,Thing,All0,All1).
modify(Place0-Place1,ThingIn,ThingOut,AllIn,AllOut):-
	modify(Place0,SomethingIn,SomethingOut,AllIn,AllOut),
	modify(Place1,ThingIn,ThingOut,SomethingIn,SomethingOut).
% funktioniert fr eine Liste
% packe Thing in die Liste 
% A=types(t1,Rest)
% pushf(types^pred,x,A,B)
% B=types(t1,[x|Rest])
pushf(Place,Thing,AllIn,AllOut):-
	modify(Place,Rest,[Thing|Rest],AllIn,AllOut).
popf(Place,Thing,AllIn,AllOut):-
	modify(Place,[Thing|Rest],Rest,AllIn,AllOut).

user:term_expansion(:-declare_struct(Name,Attrs),
		 Clauses):-
	length(Attrs,N),
	findall(struct:modify(Name^AttrName,A,B,ItA,ItB),
		(nth1(I,Attrs,AttrName),
		 genmodify(Name,N,I,A,B,ItA,ItB)),
		Clauses).
user:goal_expansion(Goal0,Goal1):-
	functor(Goal0,F,_),
	expand_inline(F),
	clause(Goal0,Goal1).

/*
%% The macro expansion, if used together with cuts,
%% can lead to unexpected effects similar to the
%% red-cut phenomenon described in Sterling/Shapiro, p. 202.
%% This should normally not be a problem, since the "after"
%% side is not instantiated before the predicate call.
%% Another case where we get wrong results is this one:
%% once(modify(foo^bar,a,b,In,Out);modify(foo^baz,c,d,In,Out))
user:goal_expansion(modify(Place,ThingIn,ThingOut,AllIn,AllOut),
		    true):-
	ground(Place),
	modify(Place,ThingIn,ThingOut,AllIn,AllOut).
*/

%% This one fixes the issue mentioned above, but is
%% somewhat less efficient because of the two additional
%% unifications.
user:goal_expansion(modify(Place,Thing0,Thing1,All0,All1),
		    (All0=All0a,All1=All1a)):-
	ground(Place),
	modify(Place,Thing0,Thing1,All0a,All1a).


emacs_prolog_colours:term_colours((:-declare_struct(_,_)),
				  expanded-[expanded-[identifier,
						      classify]]).
genmodify(Name,N,Index,A,B,ItA,ItB):-
	functor(ItA,Name,N),
	functor(ItB,Name,N),
	arg(Index,ItA,A),
	arg(Index,ItB,B),
	Index1 is Index-1,
	Index2 is Index+1,
	findall(X,between(1,Index1,X),L1),
	findall(X,between(Index2,N,X),L2),
	checklist(gm(ItA,ItB),L1),
	checklist(gm(ItA,ItB),L2).
gm(ItA,ItB,I):-
	arg(I,ItA,X),
	arg(I,ItB,X).
	
	
