:-module(transform,[nform2/2, simplifyUni/3]).

:- use_module(logging).

%% Normalform #1
%% - no equation of the form Var=Term occurs in the body

%=simplifyUni(What,TermsIn,TermsOut)
%% replaces the subgoals of the form Var=Expr in
%% <i>TermsIn</i> by substituting Var with Expr
%% the result is <i>TermsOut</i>
%% the list of simplified equations is put in <i>What</i>
simplifyUni(What,TermsIn,TermsOut):-
	logging(auswertung,'<SIMPLIFYUNI>TermsIn=~w', [TermsIn]),
	% in case that we have '<var>'('Varname')=xxxx like '<var>'('X')=2264
	select('<var>'(Varname)=Expr,TermsIn,Terms1)->
	What=['<var>'(Varname)=Expr|What2],
	subst(Varname,Expr,Terms1,Terms2),
	simplifyUni(What2,Terms2,TermsOut);
	
	% in case that we have xxxx='<var>'('Varname') in the term list
	select(Expr='<var>'(Varname),TermsIn,Terms1)->
	What=[Expr='<var>'(Varname)|What2],
	subst(Varname,Expr,Terms1,Terms2),
	simplifyUni(What2,Terms2,TermsOut);
	% in case that we have a complex structure like:
	% ausleihe('<var>'('X'), '<var>'('Y'), '<a-var>'('_405'))
	What=[],TermsOut=TermsIn,
	logging(auswertung,'<SIMPLIFYUNI>TermsOut=~w', [TermsOut]).

%= subst(+Varname,+What,+TermIn,-TermOut)
%% substitute the +Varname in +TermIn by the value +What.
%% the result is -TermOut.

subst(Varname,What,'<var>'(Varname),What):-!.
subst(_,_,A,A):-
	atomic(A),!.
subst(_,_,'<a-var>'(N),'<a-var>'(N)):-!.
subst(Varname,What,Term1,Term2):-
	Term1=..[Fun|Args1],
	maplist(subst(Varname,What),Args1,Args2),
	Term2=..[Fun|Args2].


%% Normalform #2
%% - The head of a clause contains only variables as an argument
%% - The body of a clause contains
%%   - equations with a variable on the left side
%%     and a functor with only variables as arguments on the other
%%   - predicate calls with only variables as arguments
%%
%% TODO:
%% - eliminate equations of the form Var1=Var2
%% - if A=f(C,D) and B=f(C,D), substitute A with B
%% - but first think about possible problems

%= nform2(Head-Body, Head2-Body2)
%% Transform a clause in the form Head-BodyTerms
%% e.g. odd(s(0))-[] => odd($1)-[$1=s($2),$2=0]
nform2(Head-Body,Head2-Body2):-
	nform2c(Head,Head2,Body2,B0),
	nform2d(Body,B0,[]).

%% transform a single term occuring in a value position
%% and generate a list of additional equations
%% e.g. nform(s(b),s($1))-->[$1=b]
nform2a('<var>'(X),'<var>'(X))-->!.
nform2a('<a-var>'(X),'<a-var>'(X))-->!.
nform2a(Term,'<var>'(X))-->
	{gensym('$',X),
	 Term=..[Functor|Args]},
	['<var>'(X)=Term2],
	nform2b(Args,Args2),
	{Term2=..[Functor|Args2]}.
%% transform a list of terms
nform2b([],[])-->[].
nform2b([A|As],[B|Bs])-->
	nform2a(A,B),
	nform2b(As,Bs).
%% transform a single term occuring in a call position
%% (that is, as a subgoal)
nform2c(A is B, A is B)-->!.
nform2c(Term,Term2)-->
	{Term=..[Functor|Args]},
	nform2b(Args,Args2),
	{Term2=..[Functor|Args2]}.
%% transform a list of subgoals
nform2d([])-->[].
nform2d([A|As])-->
	nform2c(A,B),
	[B],
	nform2d(As).


