:- use_module(library('http/xpce_httpd')).
:- use_module(library('http/http_client')).
:- use_module(plht).
:- use_module(logging).

:- dynamic index_file/1.
:- dynamic simple_mime/2.
:- dynamic processed_file/2.

index_file('index.html').
index_file('index.plht').
simple_mime('.html','text/html').
simple_mime('.htm','text/html').
simple_mime('.txt','text/plain').
simple_mime('.gif','image/gif').
simple_mime('.jpg','image/jpeg').
simple_mime('.png','image/png').
processed_file('.plht',process_plht).

reply(Query):-
	member(path(Path),Query),
	concat_atom(['',Filename],'/',Path),
	(   
	concat_atom([web,Filename],'/',RealPath),
	    exists_file(RealPath)->publish_file(RealPath,Query);
	Filename='',index_file(I),
	    concat_atom([web,I],'/',RealPath),
	    exists_file(RealPath)->publish_file(RealPath,Query)).

publish_file(Real_Path,_):-
	simple_mime(Ext,Mimetype),
	atom_concat(_,Ext,Real_Path),
	format('Content-type: ~a\r\n\r\n',[Mimetype]),
	open(Real_Path,read,In),
	current_output(Out),
	copy_stream_data(In,Out),
	close(In).

publish_file(Real_Path,Query):-
	processed_file(Ext,Pred),
	atom_concat(_,Ext,Real_Path),
	call(Pred,Real_Path,Query).

publish_file(_,Query):-
	member(method(post),Query)->
	format('Content-type: text/plain\r\n\r\n'),
	writeq(Query),
	write('\r\n'),
	once(member(search(Search),Query);Search=[]),
	http_read_data(Query,Data,[]),
	append(Data,Search,AllValues),
	format('~n'),
	writeq(AllValues).
	

flush_cache:-
	ignore((current_predicate(plht:PredName/_),
	       atom_concat('*plht_',_,PredName),
	       abolish(PredName/1),
	       fail)).

run_server(Port):-
	http_server(reply,[port(Port)]).
